/*
 * scantest.c: use poly_scan() for Gouraud shading and z-buffer demo.
 * Given the screen space X, Y, and Z of N-gon on command line,
 * print out all pixels during scan conversion.
 * This code could easily be modified to actually read and write pixels.
 *
 * Paul Heckbert        Dec 1989
 */

#include <stdio.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include "polygon.h"

#define XMAX 256               /* hypothetical image width */
#define YMAX 256               /* hypothetical image height */

/* random number between 0 and 1 */
#define FRANDOM() ((double) random()/INT_MAX)

static void pixelproc();

int Imagefile;
double **Zbuffer;
char **Image_RED;
char **Image_GREEN;
char **Image_BLUE;
char  *Output_Buffer;
int    Output_Buffer_len;

main(int argc, char *argv[])
{
   FILE           *InputFile;
   int             i;
   int             j;
   int             Num_p;
   polygon        *p;
   vertex         *Vertices;
   char            Header[80];
   int             Index;

   /* screen clipping window */
   static window   Window = {0, 0, XMAX - 1, YMAX - 1};

   if (argc < 2)
   {
      fprintf(stderr, "Usage: scantest inputfile\n");
      exit(1);
   }

   if ((Imagefile = open("test.ppm", O_CREAT | O_WRONLY | O_TRUNC,
                         S_IREAD | S_IWRITE)) == -1)
   {
      fprintf(stderr, "Cannot open output file.\n");
      exit(1);
   }

   if ((InputFile = fopen(argv[1], "r")) == NULL)
   {
      fprintf(stderr, "Cannot open input file -- %s.\n", argv[1]);
      exit(1);
   }

   if ((Zbuffer = (double **) malloc(XMAX * sizeof(double *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Zbuffer[0] = (double *) malloc(XMAX * YMAX * sizeof(double))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_RED = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_RED[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_GREEN = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_GREEN[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_BLUE = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_BLUE[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }
   Output_Buffer_len = 3 * YMAX * sizeof(char);
   if ((Output_Buffer = (char *) malloc(Output_Buffer_len)) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   for (i = 1; i < XMAX; i++)
   {
      Zbuffer[i] = &(Zbuffer[i - 1][YMAX]);
      Image_RED[i] = &(Image_RED[i - 1][YMAX]);
      Image_GREEN[i] = &(Image_GREEN[i - 1][YMAX]);
      Image_BLUE[i] = &(Image_BLUE[i - 1][YMAX]);
      for (j = 0; j < YMAX; j++)
      {
         Zbuffer[i][j] = HUGE_VAL;
         Image_RED[i][j] = 0;
         Image_GREEN[i][j] = 0;
         Image_BLUE[i][j] = 0;
      }
   }

   srandom(time(NULL));

   fscanf(InputFile, "%d%*c", &Num_p);

   p = new_n_polygons(Num_p, 3);

   for (i = 0; i < Num_p; i++)
   {
      fscanf(InputFile, "%d%*c", &(p[i].n));
      for (j = 0; j < p[i].n; j++)
      {
         fscanf(InputFile, "%lf %lf %lf%*c", &(p[i].Vertices[j].sx),
                &(p[i].Vertices[j].sy), &(p[i].Vertices[j].sz));
         p[i].Vertices[j].r = FRANDOM();  /* random vertex colors, for kicks */
         p[i].Vertices[j].g = FRANDOM();
         p[i].Vertices[j].b = FRANDOM();
         p[i].mask = VERTEX_MASK(sx) | VERTEX_MASK(sy) | VERTEX_MASK(sz) |
                     VERTEX_MASK(r) | VERTEX_MASK(g) | VERTEX_MASK(b);
      }
   }

   fclose(InputFile);

   for (i = 0; i < Num_p; i++)
   {
      /* interpolate sx, sy, sz, r, g, and b in poly_scan */
#if 0
      polygon_print("scan converting the polygon", &(p[i]));
#endif
      polygon_scan_convert(&(p[i]), &Window, pixelproc); /* scan convert! */
   }

   sprintf(Header, "P6\n%d %d\n255\n\0", XMAX, YMAX);
   write(Imagefile, Header, strlen(Header));

   for (i = 0; i < XMAX; i++)
   {
      for (j = 0, Index = 0; j < YMAX; j++, Index += 3)
      {
         Output_Buffer[Index] = Image_RED[i][j];
         Output_Buffer[Index + 1] = Image_GREEN[i][j];
         Output_Buffer[Index + 2] = Image_BLUE[i][j];
      }
      write(Imagefile, Output_Buffer, Output_Buffer_len);
   }

   close(Imagefile);
}

/* called at each pixel by poly_scan */
static void pixelproc(int x, int y, vertex *point)
{
#if 0
   printf("pixel (%d,%d) screenz=%g rgb=(%g,%g,%g)\n",
      x, y, point->sz, point->r, point->g, point->b);
#endif
   /*
    * in real graphics program you could read and write pixels, e.g.:
    * 
    * if (point->sz < zbuffer_read(x, y)) { image_write_rgb(x, y, point->r,
    * point->g, point->b); zbuffer_write(x, y, point->sz); }
    */

    if (point -> sz < Zbuffer[x][y])
    {
       Zbuffer[x][y] = point -> sz;
       Image_RED[x][y] = (int) floor((point->r * 255.0) + 0.5);
       Image_GREEN[x][y] = (int) floor((point->g * 255.0) + 0.5);
       Image_BLUE[x][y] = (int) floor((point->b * 255.0) + 0.5);
    }
}
